import { Database, Spanner } from '@google-cloud/spanner';
import dotenv from 'dotenv';
import * as fs from 'fs';
import { parse } from './parse';
import { dbSchema } from './schema';
import {
  ActedInRelationship,
  DirectedRelationship,
  Genre,
  HasGenreRelationship,
  Movie,
  Person
} from './types';

// Load environment variables from .env file
dotenv.config();

// Load configuration from environment variables
const config = {
  projectId: process.env.SPANNER_PROJECT_ID,
  instanceId: process.env.SPANNER_INSTANCE_ID,
  databaseId: process.env.SPANNER_DATABASE_ID,
  credentialsPath: process.env.GOOGLE_APPLICATION_CREDENTIALS
};

// Validate required environment variables
function validateConfig() {
  const requiredVars = [
    'SPANNER_PROJECT_ID',
    'SPANNER_INSTANCE_ID',
    'SPANNER_DATABASE_ID',
    'GOOGLE_APPLICATION_CREDENTIALS'
  ];

  const missingVars = requiredVars.filter(varName => !process.env[varName]);

  if (missingVars.length > 0) {
    throw new Error(
      `Missing required environment variables: ${missingVars.join(', ')}\n` +
        'Please check your .env file or environment settings.'
    );
  }

  // Verify that the files exist
  if (!fs.existsSync(config.credentialsPath!)) {
    throw new Error(
      `Credentials file not found: ${config.credentialsPath}\n` +
        'Please check your GOOGLE_APPLICATION_CREDENTIALS environment variable.'
    );
  }
}

// First pass to prepare consistent IDs

// Function to delete the database
async function dropAndCreateDatabase(drop: boolean = false): Promise<void> {
  console.log('Setting up database...');

  // Set Google credentials
  process.env.GOOGLE_APPLICATION_CREDENTIALS = config.credentialsPath;

  // Create Spanner client
  const spanner = new Spanner({
    projectId: config.projectId
  });

  // Get instance
  const instance = spanner.instance(config.instanceId!);

  // Check if database exists, and delete it if it does
  try {
    const database = instance.database(config.databaseId!);
    const [databaseExists] = await database.exists();

    if (databaseExists) {
      if (drop) {
        console.log(`Database ${config.databaseId} exists, dropping it...`);
        await database.delete();
        console.log(`Database ${config.databaseId} dropped.`);
      } else return;
    }
  } catch (error) {
    console.error('Error checking/dropping database:', error);
    throw error;
  }

  // Create new database
  try {
    console.log(`Creating new database ${config.databaseId}...`);
    const [, operation] = await instance.createDatabase(config.databaseId!, {});

    await operation.promise();
    console.log(`Database ${config.databaseId} created.`);
  } catch (error) {
    console.error('Error creating database:', error);
    throw error;
  }
}

// Function to create the database schema for graph data model
async function createDatabaseSchema(drop: boolean = false): Promise<void> {
  console.log('Creating database schema...');

  // Create Spanner client
  const spanner = new Spanner({
    projectId: config.projectId
  });

  // Get instance and database
  const instance = spanner.instance(config.instanceId!);
  const database = instance.database(config.databaseId!);

  // Check if schema already exists
  const [schema] = await database.getSchema();
  if (schema.length > 0 && !drop) {
    console.log('Schema already exists, skipping creation.');
    return;
  }

  // Create schema
  try {
    const statements = dbSchema;

    const [operation] = await database.updateSchema(statements);
    await operation.promise();

    console.log('Schema created successfully!');
  } catch (error) {
    console.error('Error creating schema:', error);
    throw error;
  }
}

// Function to process data and populate database
async function processAndPopulateData(): Promise<void> {
  console.log('Starting data processing and import...');

  // Create Spanner client
  const spanner = new Spanner({
    projectId: config.projectId
  });

  // Get instance and database
  const instance = spanner.instance(config.instanceId!);
  const database = instance.database(config.databaseId!);

  const {
    movies,
    people,
    genres,
    directedRelationships,
    actedInRelationships,
    hasGenreRelationships
  } = await parse('./imdb_movie_dataset.csv');

  // Now insert the data in the correct order
  try {
    // 1. First insert nodes
    console.log('\nInserting nodes...');

    // Insert movies
    await insertMovies(database, movies);
    console.log('Movies inserted successfully');

    // Insert people
    await insertPeople(database, people);
    console.log('People inserted successfully');

    // Insert genres
    await insertGenres(database, genres);
    console.log('Genres inserted successfully');

    // 2. Then insert relationships
    console.log('\nInserting relationships...');

    // Insert directed relationships
    await insertDirectedRelationships(database, directedRelationships);
    console.log('DIRECTED relationships inserted successfully');

    // Insert acted_in relationships
    await insertActedInRelationships(database, actedInRelationships);
    console.log('ACTED_IN relationships inserted successfully');

    // Insert has_genre relationships
    await insertHasGenreRelationships(database, hasGenreRelationships);
    console.log('HAS_GENRE relationships inserted successfully');

    console.log('\nData import completed successfully!');
  } catch (error) {
    console.error('Error inserting data:', error);
    throw error;
  }
}

// Insert movies into Spanner
async function insertMovies(
  database: Database,
  movies: Movie[]
): Promise<void> {
  const table = 'Movies';

  // Insert in batches of 100
  const batchSize = 100;
  for (let i = 0; i < movies.length; i += batchSize) {
    const batch = movies
      .slice(i, i + batchSize)
      .map(({ id, title, year, description, runtime, votes }) => {
        return { id, title, year, description, runtime, votes };
      });

    try {
      await database.table(table).insert(batch);
      console.log(
        `Inserted ${batch.length} movies (${i + batch.length}/${movies.length})`
      );
    } catch (error) {
      console.error(
        `Error inserting movies batch starting at index ${i}:`,
        error
      );
      console.error('Batch:', batch);
      throw error;
    }
  }
}

// Insert people into Spanner
async function insertPeople(
  database: Database,
  people: Person[]
): Promise<void> {
  const table = 'People';

  // Insert in batches of 100
  const batchSize = 100;
  for (let i = 0; i < people.length; i += batchSize) {
    const batch = people.slice(i, i + batchSize);

    try {
      await database.table(table).insert(batch);
      console.log(
        `Inserted ${batch.length} people (${i + batch.length}/${people.length})`
      );
    } catch (error) {
      console.error(
        `Error inserting people batch starting at index ${i}:`,
        error
      );
      throw error;
    }
  }
}

// Insert genres into Spanner
async function insertGenres(database: any, genres: Genre[]): Promise<void> {
  const table = 'Genres';

  // Insert in batches
  const batchSize = 50;
  for (let i = 0; i < genres.length; i += batchSize) {
    const batch = genres.slice(i, i + batchSize);

    try {
      await database.table(table).insert(batch);
      console.log(
        `Inserted ${batch.length} genres (${i + batch.length}/${genres.length})`
      );
    } catch (error) {
      console.error(
        `Error inserting genres batch starting at index ${i}:`,
        error
      );
      throw error;
    }
  }
}

// Insert directed relationships into Spanner
async function insertDirectedRelationships(
  database: Database,
  relationships: DirectedRelationship[]
): Promise<void> {
  const table = 'Directed';

  // Insert in batches
  const batchSize = 100;
  for (let i = 0; i < relationships.length; i += batchSize) {
    const batch = relationships.slice(i, i + batchSize);

    try {
      await database.table(table).insert(batch);
      console.log(
        `Inserted ${batch.length} directed relationships (${i + batch.length}/${
          relationships.length
        })`
      );
    } catch (error) {
      console.error(
        `Error inserting directed relationships batch starting at index ${i}:`,
        error
      );
      throw error;
    }
  }
}

// Insert acted_in relationships into Spanner
async function insertActedInRelationships(
  database: Database,
  relationships: ActedInRelationship[]
): Promise<void> {
  const table = 'ActedIn';

  // Insert in batches
  const batchSize = 100;
  for (let i = 0; i < relationships.length; i += batchSize) {
    const batch = relationships.slice(i, i + batchSize);

    try {
      await database.table(table).insert(batch);
      console.log(
        `Inserted ${batch.length} acted_in relationships (${i + batch.length}/${
          relationships.length
        })`
      );
    } catch (error) {
      console.error(
        `Error inserting acted_in relationships batch starting at index ${i}:`,
        error
      );
      throw error;
    }
  }
}

// Insert has_genre relationships into Spanner
async function insertHasGenreRelationships(
  database: Database,
  relationships: HasGenreRelationship[]
): Promise<void> {
  const table = 'HasGenre';

  // Insert in batches
  const batchSize = 100;
  for (let i = 0; i < relationships.length; i += batchSize) {
    const batch = relationships.slice(i, i + batchSize);

    try {
      await database.table(table).insert(batch);
      console.log(
        `Inserted ${batch.length} has_genre relationships (${
          i + batch.length
        }/${relationships.length})`
      );
    } catch (error) {
      console.error(
        `Error inserting has_genre relationships batch starting at index ${i}:`,
        error
      );
      throw error;
    }
  }
}

// Main execution
async function main() {
  try {
    // Validate environment variables
    validateConfig();

    // 2. Drop and recreate the database
    await dropAndCreateDatabase(true);

    // 3. Create the database schema
    await createDatabaseSchema(true);

    // 4. Process and populate the data
    await processAndPopulateData();

    console.log('\n==========================================');
    console.log('IMDB GRAPH DATABASE SETUP SUCCESSFUL!');
    console.log('==========================================');
    console.log('You can now run your Cypher queries on the database.');

    // let's try to query the database with GQL
  } catch (error) {
    console.error('Error setting up IMDB graph database:', error);
    process.exit(1);
  }
}

// Run the script
main().catch(console.error);
